#include <stdlib.h>
#include <tap.h>
#include "libFireBird.h"

TAP_ID( 0x810efaac );
#define TAPNAME "ClockTest"
TAP_PROGRAM_NAME(TAPNAME " v0.2");
TAP_AUTHOR_NAME("R2-D2");
TAP_DESCRIPTION("Test");
TAP_ETCINFO(__DATE__);

dword (*fwUTCtoLocal)(dword utc);
dword (*fwGetNow)();
dword *UTC;

dword UTCtoLocal(dword utc) { return CallFirmwareA(fwUTCtoLocal, utc); }
dword GetNow() { return CallFirmwareA(fwGetNow); }

#define WAITTICKS (10*60*100)

dword TAP_EventHandler( word event, dword param1, dword param2 )
{
  static dword tick = 0, now = 0;
  if(event == EVT_IDLE)
  {
    dword c = TAP_GetTick();
    if(tick == 0)
    {
      tick = c;
      now = GetNow();
    }
    if(c - tick > WAITTICKS)
    {
      ShowMessageWin(TAPNAME, "Completed", "No errors found in 10min sample", 500);
      TAP_Exit();
    }
    else
    {
      dword u = *UTC, t = UTCtoLocal(u);
      if((t & 0xff) != (u & 0xff))
      {
        char err[20], str[40];
        int d = (t & 0xff) - (u & 0xff);
        if(d<0)
          d = -(d + 60);
        TAP_SPrint(err, "Error at %02d:%02d", (now>>8) & 0xff, now & 0xff);
        TAP_SPrint(str, "%02d:%02d -> %02d:%02d (%+dmins diff)", (u>>8) & 0xff, u & 0xff, (t>>8) & 0xff, t & 0xff, d);
        ShowMessageWin(TAPNAME, err, str, 500);
        TAP_Exit();
      }
      else
        now = GetNow();
    }
  }
  else if(event == EVT_KEY && param1 == RKEY_Exit)
  {
    ShowMessageWin(TAPNAME, "Aborted", "(No errors found)", 500);
    TAP_Exit();
  }
  return param1;
}

void *FindFnStart(void *vp)
{
  dword *p = vp;
  while((*p & 0xfffff000) != 0x27bdf000 && (*p & CMD_MASK) != JMP_CMD)
    p--;
  return p;
}

int TAP_Main(void)
{
  // a key handler @ 80169684, subTimestamps call@+1, addTimestamps call@+5, getNowTimestamp call@+8
  dword *tm = (void*)FindInstructionSequence(
    "03c02025 0c04c5cd 24050c00 10000004 0040f025 0c04c56f 24050c00 0040f025 0c04f110 00000000",
    "fc1fffff fc000000 ffffffff ffffffff ffff07ff fc000000 ffffffff ffff07ff fc000000 ffffffff",
    0x80150000, 0x801f0000, 0, FALSE);
  // utctolocal_time @ 801318a4, ClockSetup: gp+short@+17, UTCDateTime: gp+short@+41, GetTime(&date,&hour,&min,&sec) call @+4
  short *u = (void*)FindInstructionSequence(
    "0080f025 27a40024 27a50020 27a60021 0c04ccaf 27a70022 97b90024 93b80020 8f83a11c 0019cc00 0018c200 93af0021 906e0004 0338c825 032f3825 000e7780 000e7782 24190001 55d9001b 84790000 8f8384c0",
    "ffffffff ffffffff ffffffff ffffffff fc000000 ffffffff ffffffff ffffffff ffe00000 ffffffff ffffffff ffffffff fc1fffff ffffffff ffff07ff ffffffff ffffffff ffffffff ffffff00 00000000 ffe00000",
    0x80120000, 0x801a0000, 0, FALSE);
  TAP_Print("%s: @ %p\n", TAPNAME, _tap_startAddr);
  InitTAPex();
  InitTAPAPIFix();
  if(!u || !tm)
  {
    char str[80];
//    TAP_Print("tp=%p r0=%p r1=%p\n", tp, RecSlot[0], RecSlot[1]);
    TAP_SPrint(str, "SysID=%d ApplID=%x", SysID, ApplID);
    ShowMessageWin(TAPNAME, "Unsupported firmware", str, 500);
    return 0;
  }
  UTC = (void*)(FWgp+u[41]);
  fwUTCtoLocal = FindFnStart(u);
  fwGetNow = (void*)ABS_ADDR(tm[8]);
  {
    dword u = *UTC, n = GetNow();
    TAP_Print(TAPNAME ": UTC=%p (%02d:%02d) 2loc=%p getnow=%p (%02d:%02d)\n",
      UTC, (u>>8) & 0xff, u & 0xff, fwUTCtoLocal, fwGetNow, (n>>8) & 0xff, n & 0xff);
  }
  return 1;
}
